from tcod.context import Context
from tcod.console import Console
from tcod.map import compute_fov

from message_log import Message, MessageLog
from render_functions import render_hp_bar, render_names_at_location

from floor_map import FloorMap #TODO: replace with "DungeonMap"

class Engine:
	def __init__(self, floor_map: FloorMap, intro_msg: Message = None, ui_width: int = None, ui_height: int = None):
		#events
		from event_handler import InGameHandler
		self.event_handler = InGameHandler(self)
		self.mouse_location = (0, 0)

		#map
		self.floor_map = floor_map
		self.floor_map.engine = self #references everywhere!

		#messages
		self.message_log = MessageLog()
		if intro_msg:
			self.message_log.push_message(intro_msg)

		#grab the player object
		self.player = self.floor_map.player
		self.ui_width = floor_map.width if ui_width is None else ui_width
		self.ui_height = 0 if ui_height is None else ui_height

		#kick off the render
		self.update_fov()

	def run_loop(self, context: Context, console: Console) -> None:
		while True:
			if self.event_handler.handle_events(context):
				self.handle_entities()

			self.handle_rendering(context, console)

	def handle_entities(self) -> None:
		self.update_fov() #knowing the FOV lets entities mess with it

		#all *actors* in the level
		for actor in set(self.floor_map.actors) - {self.player}:
			if actor.ai:
				actor.ai.apply()

	def handle_rendering(self, context: Context, console: Console) -> None:
		#map and all entities within
		self.floor_map.render(console)

		#UI
		render_hp_bar(
			console = console,
			x = 0,
			y = self.floor_map.height,
			current_value = self.player.fighter.current_hp,
			max_value = self.player.fighter.maximum_hp,
			total_width = self.ui_width // 2,
		)
		render_names_at_location(
			console = console,
			x = 1,
			y = self.floor_map.height + 2,
			engine = self,
		)
		self.message_log.render(
			console=console,
			x=self.ui_width // 2,
			y=self.floor_map.height,
			width = self.ui_width // 2,
			height = self.ui_height,
		)

		self.event_handler.render(console)

		#send to the screen
		context.present(console)
		console.clear()

	#utils
	def update_fov(self):
		self.floor_map.visible[:] = compute_fov(
			self.floor_map.tiles["transparent"],
			(self.player.x, self.player.y),
			radius = 8,
		)

		#add the visible tiles to the explored list
		self.floor_map.explored |= self.floor_map.visible
